﻿//  
//  MyNesSDL.cs
//  
//  Author:
//       Ala Hadid <ahdsoftwares@hotmail.com>
// 
//  Copyright (c) Ala Hadid 2011 
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;

using SdlDotNet;
using SdlDotNet.Graphics;
using SdlDotNet.Graphics.Sprites;
using SdlDotNet.Core;
using SdlDotNet.Input;
using SdlDotNet.Graphics.Primitives;

using MyNes.Nes;

namespace MyNesSDL
{
    public class MyNesSDL
    {
        public static NesSystem nes;
	    static string StateDirectory=".\\statesaves";
		static string SnapshotsDirectory=".\\snapshots";
		static bool fullscreen = false;
		static bool autoDetectTvSystem = true;
	    static bool resizable = true;
		static RegionFormat tvFormat= RegionFormat.NTSC;
        static string FileName = "";
		static int width = 512;
		static int height = 448;
	    static InputManager inputManager = new InputManager(new Joypad(), new Joypad(),new Joypad(), new Joypad());
		static int stateIndex = 1;
		static bool SoundOn = true;
		static bool NoLimiter = false;
		//KEYS
		static Key PauseKey = Key.F2;
		static Key SoftReset = Key.F3;	
		static Key TakeSnap = Key.F5;
	    static Key StateSave = Key.F6;
		static Key StateLoad = Key.F9;
		static Key NoLimiterKey = Key.F7;
		static Key ToggleFullscreen = Key.F12;

		public  MyNesSDL(string[] Args)
		{
			if (Args.Length > 0)
			    FileName=Args[0];
			else
			    FileName = "test.nes"; 
			#region commandlines
			for (int i = 1 ; i < Args.Length; i++)
			{
				if (!Args[i].ToString().Contains("/"))
				{
					FileName+= " " + Args[i]; 
					continue; 
				}
				//general commands
				switch(Args[i])
				{
				    case "/fullscreen":
					fullscreen = true;
					break;
				    case "/tv_auto": 
					autoDetectTvSystem = true;
					break;
			     	case "/tv_pal": 
					tvFormat= RegionFormat.PAL;
					break;
					case "/tv_ntsc":
					tvFormat= RegionFormat.NTSC;
					break;
					case "/resizable": 
					resizable = true;
					break;
					case "/statefolder": 
					i++;
					StateDirectory=Args[i];
					break;
					case "/snapfolder": 
					i++;
					SnapshotsDirectory=Args[i];
					break;
					case "/slot": 
					i++;
					stateIndex=int.Parse(Args[i]);
					Console.WriteLine("State slot "+stateIndex + " selected");
					break;
			    	case "/sound_off":  SoundOn=false;	break;
					case "/nolimiter":  NoLimiter=true;	break;
				}
				//special for window size
				if (Args[i].ToString().Length >= 2)
				{
			    	switch(Args[i].ToString().Substring(0,2))
			    	{
				       case "/w":
						width= Convert.ToInt32(Args[i].ToString().Substring(2,Args[i].ToString().Length-2)); 
						break;
			    	   case "/h":
						height= Convert.ToInt32(Args[i].ToString().Substring(2,Args[i].ToString().Length-2)); 
						break;
			    	}
				}
				#region key commands
				if (Args[i].ToString().Contains("="))
				{
					string[] cm=Args[i].ToString().Split(new char[]{'='});
					switch (cm[0])
					{
				    	case "/p1_a":
						//we'll use this loop to detect key just to make 
						//it easy for the user to send key, so no need to
						//case sensitive
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_A=key;
								break;
							}
						}
						break;
						case "/p1_b":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_B=key;
								break;
							}
						}
						break;
						case "/p1_st":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_ST=key;
								break;
							}
						}
						break;
						case "/p1_se":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_SE=key;
								break;
							}
						}
						break;
						case "/p1_up":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_UP=key;
								break;
							}
						}
						break;
						case "/p1_do":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_DO=key;
								break;
							}
						}
						break;
						case "/p1_lf":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_LF=key;
								break;
							}
						}
						break;
						case "/p1_rt":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p1_RT=key;
								break;
							}
						}
						break;

					    case "/p2_a":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_A=key;
								break;
							}
						}
						break;
						case "/p2_b":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_B=key;
								break;
							}
						}
						break;
						case "/p2_st":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_ST=key;
								break;
							}
						}
						break;
						case "/p2_se":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_SE=key;
								break;
							}
						}
						break;
						case "/p2_up":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_UP=key;
								break;
							}
						}
						break;
						case "/p2_do":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_DO=key;
								break;
							}
						}
						break;
						case "/p2_lf":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_LF=key;
								break;
							}
						}
						break;
						case "/p2_rt":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p2_RT=key;
								break;
							}
						}
						break;

						case "/p3_a":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_A=key;
								break;
							}
						}
						break;
						case "/p3_b":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_B=key;
								break;
							}
						}
						break;
						case "/p3_st":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_ST=key;
								break;
							}
						}
						break;
						case "/p3_se":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_SE=key;
								break;
							}
						}
						break;
						case "/p3_up":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_UP=key;
								break;
							}
						}
						break;
						case "/p3_do":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_DO=key;
								break;
							}
						}
						break;
						case "/p3_lf":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_LF=key;
								break;
							}
						}
						break;
						case "/p3_rt":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p3_RT=key;
								break;
							}
						}
						break;

					    case "/p4_a":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_A=key;
								break;
							}
						}
						break;
						case "/p4_b":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_B=key;
								break;
							}
						}
						break;
						case "/p4_st":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_ST=key;
								break;
							}
						}
						break;
						case "/p4_se":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_SE=key;
								break;
							}
						}
						break;
						case "/p4_up":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_UP=key;
								break;
							}
						}
						break;
						case "/p4_do":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_DO=key;
								break;
							}
						}
						break;
						case "/p4_lf":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_LF=key;
								break;
							}
						}
						break;
						case "/p4_rt":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								inputManager.p4_RT=key;
								break;
							}
						}
						break;

						case "/key_pause":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								PauseKey=key;
								break;
							}
						}
						break;
						case "/key_softreset":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								SoftReset=key;
								break;
							}
						}
						break;
						case "/key_takesnap":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								TakeSnap=key;
								break;
							}
						}
						break;
						case "/key_statesave":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								StateSave=key;
								break;
							}
						}
						break;
						case "/key_stateload":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								StateLoad=key;
								break;
							}
						}
						break;
						case "/key_nolimiter":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								NoLimiterKey=key;
								break;
							}
						}
						break;
						case "/key_fullscreen":
						foreach (Key key in Enum.GetValues(typeof(Key)))
						{
							if (key.ToString().ToLower()==cm[1].ToString().ToLower())
							{
								ToggleFullscreen=key;
								break;
							}
						}
						break;
					}
				}
				#endregion
			}
			#endregion
			
			CONSOLE.DebugRised+=new EventHandler<DebugArg>(DebugRised);
			
			//create directories
			Directory.CreateDirectory(StateDirectory);
			Directory.CreateDirectory(SnapshotsDirectory);
			    
			if (FileName=="" | !File.Exists(FileName))
			{	     
	    		CONSOLE.WriteLine(null, "Rom path is not exist !!", DebugStatus.Error);
				CONSOLE.WriteLine(null, "At path = " + FileName, DebugStatus.Error);
			}
			else
			{
				EnableEvents();
				LoadCart();
			}
		}
		static void DebugRised(object sender, DebugArg e)
        {
            switch (e.Status)
			{
			case DebugStatus.Cool: Console.ForegroundColor= ConsoleColor.Green; break;
			case DebugStatus.Error: Console.ForegroundColor= ConsoleColor.Red; break;
			case DebugStatus.Warning: Console.ForegroundColor= ConsoleColor.Yellow; break;
			}
			Console.WriteLine(e.DebugLine);
			Console.ResetColor();
        }
        static void EnableEvents()
        {
            Events.KeyboardDown +=
                new EventHandler<KeyboardEventArgs>(KeyboardDown);
            Events.Quit += new EventHandler<QuitEventArgs>(Quit);
			Events.VideoResize+=new EventHandler<VideoResizeEventArgs>(VideoResize);
        }
        #region EventHandler Methods
        static void VideoResize(object sender, VideoResizeEventArgs e)
        {
            ((videoSDL)nes.graphicsDevice).Resize(fullscreen,resizable,e.Width,e.Height);
        }
		static void KeyboardDown(object sender, KeyboardEventArgs e)
        {
			//exit
            if (e.Key == Key.Escape)
            {	
				nes.ShutDown();	
            }
		    if (e.Key == ToggleFullscreen)
            {	
				fullscreen = !fullscreen;
				((videoSDL)nes.graphicsDevice).Resize(fullscreen,resizable,width,height);
            }
			//soft reset
			if (e.Key == SoftReset)
			{
				nes.Cpu.SoftReset();
			}
			//toggle pause
			if (e.Key == PauseKey)
			{
	            nes.PAUSE = !nes.PAUSE;
			}
			//take snap
			if (e.Key == TakeSnap)
			{
				int i=0;
				string snapFile=SnapshotsDirectory+"\\"+
					Path.GetFileNameWithoutExtension(FileName)+"_"+i+".bmp";
				while (File.Exists(snapFile))
				{
					i++;
					snapFile=SnapshotsDirectory+"\\"+
					Path.GetFileNameWithoutExtension(FileName)+"_"+i+".bmp";
				}
	           nes.graphicsDevice.TakeSnapshot(snapFile, "");
			}
			//state
			if (e.Key == StateSave)
			{
				string stateFile=StateDirectory+"\\"+
					Path.GetFileNameWithoutExtension(FileName)+"_"+stateIndex+".mns";
				try
				{
					nes.SaveStateRequest(stateFile,
						new FileStream(stateFile,FileMode.Create,FileAccess.Write));
				}
				catch{}
			}
		    if (e.Key == StateLoad)
			{
				string stateFile=StateDirectory+"\\"+
					Path.GetFileNameWithoutExtension(FileName)+"_"+stateIndex+".mns";
				try
				{
					nes.LoadStateRequest(stateFile,
						new FileStream(stateFile,FileMode.Open,FileAccess.Read));
				}
				catch{}
			}
		    if (e.Key == NoLimiterKey)
			{
				nes.NoLimiter = !nes.NoLimiter;
				Console.WriteLine("No limiter = "+nes.NoLimiter .ToString());
			}
			switch (e.Key)
			{
			        case Key.Keypad1:stateIndex=1;Console.WriteLine("State Slot 1 selected");break;
					case Key.Keypad2:stateIndex=2;Console.WriteLine("State Slot 2 selected");break;
					case Key.Keypad3:stateIndex=3;Console.WriteLine("State Slot 3 selected");break;
					case Key.Keypad4:stateIndex=4;Console.WriteLine("State Slot 4 selected");break;
					case Key.Keypad5:stateIndex=5;Console.WriteLine("State Slot 5 selected");break;
					case Key.Keypad6:stateIndex=6;Console.WriteLine("State Slot 6 selected");break;
					case Key.Keypad7:stateIndex=7;Console.WriteLine("State Slot 7 selected");break;
					case Key.Keypad8:stateIndex=8;Console.WriteLine("State Slot 8 selected");break;
					case Key.Keypad9:stateIndex=9;Console.WriteLine("State Slot 9 selected");break;
			}
        }
        static void Quit(object sender, QuitEventArgs e)
        {
			nes.ShutDown();
            //Events.QuitApplication();
        }
        #endregion EventHandler Methods
        static void LoadCart()
        {
            //check out the rom
            NesCartridge cart = new NesCartridge(null);
            LoadRomStatus status = cart.Load(FileName, new FileStream(FileName, FileMode.Open, FileAccess.Read), null, true);
            switch (status)
            {
                case LoadRomStatus.InvalidHeader:
				CONSOLE.WriteLine(null,"Can't open this rom, invalid INES header", DebugStatus.Error);return;
                case LoadRomStatus.InvalidMapper:
				CONSOLE.WriteLine(null,"Can't open this rom, unsopported mapper", DebugStatus.Error);return;
                case LoadRomStatus.LoadFailure: 
				CONSOLE.WriteLine(null,"Can't open this rom, unknown error.", DebugStatus.Error);return;
            }
			//options 
			if (autoDetectTvSystem)
			{
	            if(cart.IsPAL)
				{
			        tvFormat= RegionFormat.PAL;	
				}
				else
				{
					tvFormat= RegionFormat.NTSC;
				}
			}
            //run the nes
			Console.WriteLine("Creating new nes engine ...");
            nes = new NesSystem(new Timer());
			//sram stream
			nes.Cartridge.SavePath = FileName.Substring(0, FileName.Length - 3) + "sav";
			Stream sramStream = Stream.Null;
            if (File.Exists(nes.Cartridge.SavePath))
                sramStream = new FileStream(nes.Cartridge.SavePath, FileMode.Open, FileAccess.ReadWrite);
            nes.LoadRom(FileName, new FileStream(FileName, FileMode.Open, FileAccess.Read), sramStream);


            //setup

			//controls
            Console.WriteLine("Setup controls ...");
			//TODO: add support for player 3 and 4
            nes.SetupInput(inputManager,inputManager.joypad1, inputManager.joypad2,
			               inputManager.joypad3,inputManager.joypad4);
			if (nes.Cartridge.IsVSUnisystem)
				nes.CpuMemory.VSunisystemDIP=inputManager.vsunisystem= new VSUnisystemSDL();

			//video
	        Console.WriteLine("Setup video ...");
			Console.WriteLine(tvFormat.ToString()+" tv system selected");
            nes.SetupVideo( new videoSDL(tvFormat,fullscreen,resizable,width, height,
				Path.GetFileName(FileName)));


		    //sound
			Console.WriteLine("Setup sound ...");
            nes.SetupAudio(new soundSDL(nes.Apu, SoundOn));
			nes.NoLimiter=NoLimiter;

            //turn on
			nes.TurnOn();
			//settings
			nes.SetTVFormat(tvFormat, new PaletteFormat());

			//run
			Console.WriteLine("Run !");
			nes.Execute();
		    Events.Close();
        }
    }
}
