﻿//  
//  videoSDL.cs
//  
//  Author:
//       Ala Hadid <ahdsoftwares@hotmail.com>
// 
//  Copyright (c) Ala Hadid 2011 
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Runtime.InteropServices;
using MyNes.Nes.Output.Video;
using MyNes.Nes;

using SdlDotNet;
using SdlDotNet.Graphics;
using SdlDotNet.Graphics.Sprites;
using SdlDotNet.Core;
using SdlDotNet.Input;
using SdlDotNet.Graphics.Primitives;

namespace MyNesSDL
{
   unsafe class videoSDL:IVideoDevice
    {
		//make 2 surfaces for resizing and fullscreen
        private Surface screen; 
        private Surface screen_back;
		string romName="";
        int* pointer;
		int firstToCut=0;
		int dtbefore=DateTime.Now.Second;
		
        public videoSDL(RegionFormat tv, bool fullScreen,bool resizable, int w, int h, string romName)
        {
			this.romName=romName;
            Video.WindowIcon();
            Video.WindowCaption = romName + " - My Nes";
		
			Resize(fullScreen,resizable,  w,  h);
			
			if (tv== RegionFormat.NTSC)
			{
                screen_back = new Surface(256, 224, 32);
				firstToCut=8;
			}
			else
			{
		        screen_back = new Surface(256, 240, 32);
			}
            //screen.AlphaBlending = false;
           // screen_back.AlphaBlending = false;
            pointer = (int*)screen_back.Pixels;
        }
		public void Resize(bool fullScreen,bool resizable, int w, int h)
		{		
			if (fullScreen)
				resizable = false;
			screen = Video.SetVideoMode(w, h, 32, resizable, false, fullScreen);
		}
        public string Name
        {
            get { return "SDL video for My Nes"; }
        }

        public string Desc
        {
            get { return ""; }
        }

        public void Begin()
        {
        
        }

        public void BlankPixel(int X, int Y)
        {
       
        }

        public void RenderFrame(int[] ScreenBuffer)
        {
			for (int i = (256 * firstToCut); i < ScreenBuffer.Length -
				(256 * firstToCut); i++)
            {
                pointer[i - (256 * firstToCut)] = ScreenBuffer[i];
            }
			 //Marshal.Copy(ScreenBuffer, (256 * firstToCut), pointer, 256 * 224);
             screen.Blit(screen_back.CreateStretchedSurface(screen.Size));
			//calculate fps
			if (DateTime.Now.Second-dtbefore>=1)
			{
				Video.WindowCaption = romName + " - My Nes ("+MyNesSDL.nes.FPS+" FPS)";
				MyNesSDL.nes.FPS=0;
			}
			 dtbefore= DateTime.Now.Second;
             screen.Update();
            //check for events
           // while (Events.Poll())
            //{ }
			Events.Poll();
        }

        public void TakeSnapshot(string SnapPath, string Format)
        {
            screen_back.SaveBmp(SnapPath);
			Console.WriteLine("Snapshot taken");
        }

        public void DrawText(string Text)
        {

        }

        public void Clear()
        {
         
        }

        public void UpdateSize(int X, int Y, int W, int H)
        {
   
        }

        public bool IsSizable
        {
            get { return false; }
        }

        public bool IsRendering
        {
            get { return false; }
        }

        public bool CanRender
        {
            get
            {
                return true;
            }
            set
            {
         
            }
        }

        public bool FullScreen
        {
            get
            {
                return false;
            }
            set
            {
             
            }
        }
		public void Dispose()
		{
			Shutdown();
		}
        public bool SupportFullScreen
        {
            get { return true; }
        }

        public void Shutdown()
        {
           	Events.CloseVideo();
        }

        public void ChangeSettings()
        {
     
        }
    }
}
